import os
from textwrap import dedent
from unittest import (
    TestCase,
    mock,
)

import pcs.lib.booth.status as lib
from pcs import settings
from pcs.common import (
    file_type_codes,
    reports,
)
from pcs.common.reports import ReportItemSeverity as Severities
from pcs.common.reports import codes as report_codes
from pcs.lib.booth import constants
from pcs.lib.external import CommandRunner

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_raise_library_error
from pcs_test.tools.command_env import get_env_tools


class GetDaemonStatusTest(TestCase):
    def setUp(self):
        self.mock_run = mock.MagicMock(spec_set=CommandRunner)

    def test_no_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_daemon_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status"]
        )

    def test_with_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_daemon_status(self.mock_run, "name"))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status", "-c", "name"]
        )

    def test_daemon_not_running(self):
        self.mock_run.run.return_value = ("", "error", 7)
        self.assertEqual("", lib.get_daemon_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status"]
        )

    def test_failure(self):
        self.mock_run.run.return_value = ("out", "error", 1)
        assert_raise_library_error(
            lambda: lib.get_daemon_status(self.mock_run),
            (
                Severities.ERROR,
                report_codes.BOOTH_DAEMON_STATUS_ERROR,
                {"reason": "error\nout"},
            ),
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "status"]
        )


class GetTicketsStatusTest(TestCase):
    def setUp(self):
        self.mock_run = mock.MagicMock(spec_set=CommandRunner)

    def test_no_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_tickets_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "list"]
        )

    def test_with_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual(
            "output", lib.get_tickets_status(self.mock_run, "name")
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "list", "-c", "name"]
        )

    def test_failure(self):
        self.mock_run.run.return_value = ("out", "error", 1)
        assert_raise_library_error(
            lambda: lib.get_tickets_status(self.mock_run),
            (
                Severities.ERROR,
                report_codes.BOOTH_TICKET_STATUS_ERROR,
                {"reason": "error\nout"},
            ),
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "list"]
        )


class GetPeersStatusTest(TestCase):
    def setUp(self):
        self.mock_run = mock.MagicMock(spec_set=CommandRunner)

    def test_no_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_peers_status(self.mock_run))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "peers"]
        )

    def test_with_name(self):
        self.mock_run.run.return_value = ("output", "", 0)
        self.assertEqual("output", lib.get_peers_status(self.mock_run, "name"))
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "peers", "-c", "name"]
        )

    def test_failure(self):
        self.mock_run.run.return_value = ("out", "error", 1)
        assert_raise_library_error(
            lambda: lib.get_peers_status(self.mock_run),
            (
                Severities.ERROR,
                report_codes.BOOTH_PEERS_STATUS_ERROR,
                {"reason": "error\nout"},
            ),
        )
        self.mock_run.run.assert_called_once_with(
            [settings.booth_binary, "peers"]
        )


class CheckAuthfileMisconfiguration(TestCase):
    def setUp(self):
        self.instance_name = "instance_name"
        self.config_path = os.path.join(
            settings.booth_config_dir, f"{self.instance_name}.conf"
        )
        self.env_assist, self.config = get_env_tools(self)

    def _run(self):
        env = self.env_assist.get_env()
        return lib.check_authfile_misconfiguration(
            env.get_booth_env(self.instance_name), env.report_processor
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_nothing_enabled(self):
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_file_doesnt_exist_set_enabled(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            exists=False,
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_file_doesnt_exist_unset_enabled(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            exists=False,
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_file_doesnt_exist_both_enabled(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            exists=False,
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_set_enabled_no_authfile(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                """
                site = 1.1.1.1
                """
            ).encode("utf-8"),
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_set_enabled_authfile_enable_missing(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                """
                site = 1.1.1.1
                authfile = /path/to/authfile
                """
            ).encode("utf-8"),
        )
        self.assertEqual(
            self._run(),
            reports.messages.BoothAuthfileNotUsed(self.instance_name),
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_set_enabled_authfile_disabled(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                f"""
                site = 1.1.1.1
                authfile = /path/to/authfile
                {constants.AUTHFILE_FIX_OPTION} = no
                """
            ).encode("utf-8"),
        )
        self.assertEqual(
            self._run(),
            reports.messages.BoothAuthfileNotUsed(self.instance_name),
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_set_enabled_authfile_enabled(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                f"""
                site = 1.1.1.1
                authfile = /path/to/authfile
                {constants.AUTHFILE_FIX_OPTION} = on
                """
            ).encode("utf-8"),
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_set_and_unset_enabled(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                f"""
                site = 1.1.1.1
                authfile = /path/to/authfile
                {constants.AUTHFILE_FIX_OPTION} = yes
                """
            ).encode("utf-8"),
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_unset_enabled_option_missing(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                """
                site = 1.1.1.1
                authfile = /path/to/authfile
                """
            ).encode("utf-8"),
        )
        self.assertIsNone(self._run())

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", False)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", True)
    def test_unset_enabled_authfile_enable_present(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                f"""
                site = 1.1.1.1
                authfile = /path/to/authfile
                {constants.AUTHFILE_FIX_OPTION} = 0
                """
            ).encode("utf-8"),
        )
        self.assertEqual(
            self._run(),
            reports.messages.BoothUnsupportedOptionEnableAuthfile(
                self.instance_name
            ),
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_config_read_failure(self):
        read_failure_reason = "read failed"
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            exception_msg=read_failure_reason,
        )
        self.assertIsNone(self._run())
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.FILE_IO_ERROR,
                    file_type_code=file_type_codes.BOOTH_CONFIG,
                    operation="read",
                    reason=read_failure_reason,
                    file_path=self.config_path,
                )
            ]
        )

    @mock.patch("pcs.settings.booth_enable_authfile_set_enabled", True)
    @mock.patch("pcs.settings.booth_enable_authfile_unset_enabled", False)
    def test_config_parsing_failed(self):
        self.config.raw_file.exists(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
        )
        self.config.raw_file.read(
            file_type_codes.BOOTH_CONFIG,
            self.config_path,
            content=dedent(
                """
                site = 1.1.1.1
                invalid ' option
                """
            ).encode("utf-8"),
        )
        self.assertIsNone(self._run())
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    reports.codes.BOOTH_CONFIG_UNEXPECTED_LINES,
                    line_list=["invalid ' option"],
                    file_path=self.config_path,
                )
            ]
        )



import logging
import multiprocessing as mp
from collections import Counter
from dataclasses import dataclass
from datetime import datetime
from queue import Queue
from unittest import (
    TestCase,
    mock,
)

from tornado.testing import AsyncTestCase

from pcs.common.async_tasks.dto import (
    CommandDto,
    CommandOptionsDto,
)
from pcs.common.async_tasks.types import TaskState
from pcs.common.reports.item import ReportItemMessage
from pcs.common.reports.types import MessageCode
from pcs.daemon.async_tasks import scheduler
from pcs.daemon.async_tasks.task import TaskConfig
from pcs.daemon.async_tasks.types import Command
from pcs.lib.auth.types import AuthUser

DATETIME_NOW = datetime(2020, 2, 20, 20, 20, 20, 20)
AUTH_USER = AuthUser("username", ("group1", "group2"))
ANOTHER_AUTH_USER = AuthUser("different user name", ["no group"])


class PermissionsCheckerMock:
    def __init__(self, permissions):
        self._permissions = permissions

    def is_authorized(self, auth_user, access):
        return access not in self._permissions.get(auth_user, tuple())


@dataclass(frozen=True)
class StubReportItem(ReportItemMessage):
    """Use this class to test report handling"""

    _code = MessageCode("REPORT_STUB")

    @property
    def message(self) -> str:
        return "This is a report item used for testing."


class SchedulerTestWrapper:
    """
    Scheduler for testing

    This class contains a scheduler that is suitable for use in tier0 tests -
    - it doesn't create new processes and contains mocks for easy testing, it
    should be used with AsyncTestCase or TestCase
    """

    scheduler = None
    worker_com = None
    mp_pool_mock = None
    logger_mock = None
    logging_queue = None

    def prepare_scheduler(self):
        # Instance attributes are not created in the mock, this includes handler
        # list which is needed by QueueListener
        self.logger_mock = mock.patch(
            "pcs.daemon.async_tasks.scheduler.pcsd_logger",
            spec=logging.Logger,
            handlers=[],
        ).start()
        # We can patch Queue here because it is NOT shared between tests
        # self.worker_com = mp.Queue()
        self.worker_com = Queue()
        self.logging_queue = Queue()
        # Manager has to be mocked because it creates a new process
        # There are two queue calls, first is for worker message queue, second
        # is for the logging queue
        mock.patch(
            "multiprocessing.Manager"
        ).start().return_value.Queue.side_effect = [
            self.worker_com,
            self.logging_queue,
            # mp.Queue(),
        ]
        self.mp_pool_mock = (
            mock.patch("multiprocessing.Pool", spec=mp.Pool)
            .start()
            .return_value
        ) = mock.Mock()
        # This might be needed when logger is called by get_logger, but is it?
        self.scheduler = scheduler.Scheduler(
            scheduler.SchedulerConfig(
                worker_count=1,
                worker_reset_limit=2,
                task_config=TaskConfig(deletion_timeout=0),
            )
        )

    def _create_tasks(self, count, start_from=0):
        """Creates tasks with task_ident from id0 to idN"""
        for i in range(start_from, count + start_from):
            with mock.patch(
                "pcs.daemon.async_tasks.scheduler.get_unique_uuid"
            ) as mock_uuid:
                mock_uuid.return_value = f"id{i}"
                self.scheduler.new_task(
                    Command(
                        CommandDto(
                            f"command {i}",
                            {},
                            CommandOptionsDto(request_timeout=None),
                        )
                    ),
                    AUTH_USER,
                )


class SchedulerBaseTestCase(SchedulerTestWrapper, TestCase):
    def setUp(self):
        SchedulerTestWrapper.prepare_scheduler(self)
        super().setUp()
        self.addCleanup(mock.patch.stopall)


class SchedulerBaseAsyncTestCase(SchedulerTestWrapper, AsyncTestCase):
    def setUp(self):
        SchedulerTestWrapper.prepare_scheduler(self)
        super().setUp()
        self.addCleanup(mock.patch.stopall)


class AssertTaskStatesMixin:
    def assert_task_state_counts_equal(
        self, created, queued, executed, finished
    ):
        # pylint: disable=protected-access
        # Cannot use public method get_task because it deletes finished tasks
        state_counts = Counter(
            [task.state for task in self.scheduler._task_register.values()]
        )
        self.assertEqual(
            created,
            state_counts[TaskState.CREATED],
            "Expected {0} CREATED tasks.".format(
                state_counts[TaskState.CREATED]
            ),
        )
        self.assertEqual(
            queued,
            state_counts[TaskState.QUEUED],
            "Expected {0} QUEUED tasks.".format(state_counts[TaskState.QUEUED]),
        )
        self.assertEqual(
            executed,
            state_counts[TaskState.EXECUTED],
            "Expected {0} EXECUTED tasks.".format(
                state_counts[TaskState.EXECUTED]
            ),
        )
        self.assertEqual(
            finished,
            state_counts[TaskState.FINISHED],
            "Expected {0} FINISHED tasks.".format(
                state_counts[TaskState.FINISHED]
            ),
        )


class MockDateTimeNowMixin:
    """Capability to set up a mock of datetime.now() function in a TestCase"""

    def _init_mock_datetime_now(self):
        # Trying to shorten this is a really funny game
        # We can't patch datetime.datetime.now because builtins are immutable
        mock_datetime = mock.patch("datetime.datetime")
        self.addCleanup(mock_datetime.stop)
        mock_datetime = mock_datetime.start()
        mock_datetime.now = mock.Mock()
        mock_datetime.now.return_value = DATETIME_NOW
        return mock_datetime.now


class MockOsKillMixin:
    """Capability to set up a mock of os.kill() function in a TestCase"""

    def _init_mock_os_kill(self):
        mock_os_kill = mock.patch("os.kill")
        self.addCleanup(mock_os_kill.stop)
        return mock_os_kill.start()



from typing import Optional

from tornado.locks import Lock

from pcs.daemon import ruby_pcsd
from pcs.daemon.app.auth import (
    LegacyTokenAuthenticationHandler,
    TokenAuthProvider,
)
from pcs.daemon.app.sinatra_common import SinatraMixin
from pcs.daemon.http_server import HttpsServerManage
from pcs.lib.auth.provider import AuthProvider
from pcs.lib.auth.types import AuthUser


class SinatraRemote(LegacyTokenAuthenticationHandler, SinatraMixin):
    """
    SinatraRemote is handler for urls which should be directed to the Sinatra
    remote (non-GUI) functions.
    """

    _token_auth_provider: TokenAuthProvider
    _auth_user: Optional[AuthUser]

    def initialize(
        self, ruby_pcsd_wrapper: ruby_pcsd.Wrapper, auth_provider: AuthProvider
    ) -> None:
        # pylint: disable=arguments-differ
        super().initialize(auth_provider)
        self.initialize_sinatra(ruby_pcsd_wrapper)

    async def _handle_request(self):
        result = await self.ruby_pcsd_wrapper.request(
            self.auth_user, self.request
        )
        self.send_sinatra_result(result)


class SyncConfigMutualExclusive(SinatraRemote):
    """
    SyncConfigMutualExclusive handles urls which should be directed to the
    Sinatra remote (non-GUI) functions that can not run at the same time as
    config synchronization. The exclusivity is achieved by sync_config_lock.
    """

    __sync_config_lock: Lock

    def initialize(
        self,
        ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
        auth_provider: AuthProvider,
        sync_config_lock: Lock,
    ) -> None:
        # pylint: disable=arguments-differ
        super().initialize(ruby_pcsd_wrapper, auth_provider)
        self.__sync_config_lock = sync_config_lock

    async def _handle_request(self):
        async with self.__sync_config_lock:
            await super()._handle_request()


class SetCerts(SinatraRemote):
    """
    SetCerts handles url for setting new certificate and key. It calls the
    Sinatra for setting certificate and key and in the case of the success it
    will take care of notify of http sever about this change.
    """

    __https_server_manage: HttpsServerManage

    def initialize(
        self,
        ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
        auth_provider: AuthProvider,
        https_server_manage: HttpsServerManage,
    ):
        # pylint: disable=arguments-differ
        super().initialize(ruby_pcsd_wrapper, auth_provider)
        self.__https_server_manage = https_server_manage

    async def _handle_request(self):
        result = await self.ruby_pcsd_wrapper.request(
            self.auth_user, self.request
        )
        if result.status == 200:
            self.__https_server_manage.reload_certs()
        self.send_sinatra_result(result)


def get_routes(
    ruby_pcsd_wrapper: ruby_pcsd.Wrapper,
    sync_config_lock: Lock,
    https_server_manage: HttpsServerManage,
    auth_provider: AuthProvider,
):
    sinatra_remote_options = dict(
        ruby_pcsd_wrapper=ruby_pcsd_wrapper, auth_provider=auth_provider
    )

    return [
        # Urls protected by tokens. It is still done by ruby pcsd.
        (r"/run_pcs", SinatraRemote, sinatra_remote_options),
        (
            r"/remote/set_certs",
            SetCerts,
            dict(
                **sinatra_remote_options,
                https_server_manage=https_server_manage,
            ),
        ),
        (
            r"/remote/(set_sync_options|set_configs)",
            SyncConfigMutualExclusive,
            dict(**sinatra_remote_options, sync_config_lock=sync_config_lock),
        ),
        (r"/remote/.*", SinatraRemote, sinatra_remote_options),
    ]



from pcs.daemon import ruby_pcsd


class SinatraMixin:
    """
    Sinatra is base class for handlers which calls the Sinatra via wrapper.
    It accept ruby wrapper during initialization. It also provides method for
    transformation result from sinatra to http response.
    """

    __ruby_pcsd_wrapper: ruby_pcsd.Wrapper

    def initialize_sinatra(self, ruby_pcsd_wrapper: ruby_pcsd.Wrapper):
        self.__ruby_pcsd_wrapper = ruby_pcsd_wrapper

    def send_sinatra_result(self, result: ruby_pcsd.SinatraResult):
        for name, value in result.headers.items():
            self.set_header(name, value)
        # make sure that security related headers, which need to be present in
        # all responses, are not overridden by sinatra
        self.set_default_headers()
        self.set_status(result.status)
        self.write(result.body)

    @property
    def ruby_pcsd_wrapper(self):
        return self.__ruby_pcsd_wrapper



from functools import partial
from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.lib.cib.resource import primitive
from pcs.lib.cib.tools import IdProvider
from pcs.lib.resource_agent import ResourceAgentName

from pcs_test.tools.assertions import assert_xml_equal


class FindPrimitivesByAgent(TestCase):
    # pylint: disable=protected-access
    def setUp(self):
        self.resources_section = etree.fromstring(
            """
        <resources>
            <primitive
                class="standard" provider="provider" type="agent_type" id="r0"
            />
            <primitive
                class="something" provider="provider" type="agent_type" id="r23"
            />
            <primitive class="stonith" type="agent_type" id="r1"/>
            <primitive
                class="standard" provider="provider" type="dummy1" id="r123"
            />
            <group>
                <primitive class="stonith" type="agent_type" id="r2"/>
                <primitive
                    class="standard" provider="pacemaker" type="agent_type"
                    id="r3"
                />
                <primitive
                    class="standard" provider="provider" type="agent_type"
                    id="r4"
                />
            </group>
            <clone>
                <group>
                    <primitive
                        class="standard" provider="provider" type="agent_type"
                        id="r5"
                    />
                </group>
            </clone>
            <clone>
                <primitive
                    class="standard" provider="provider" type="agent_type"
                    id="r6"
                />
            </clone>
        </resources>
        """
        )

    def test_stonith(self):
        results = primitive._find_primitives_by_agent(
            self.resources_section,
            ResourceAgentName(
                "stonith",
                None,
                "agent_type",
            ),
        )
        expected_results = [
            '<primitive class="stonith" type="agent_type" id="r1"/>',
            '<primitive class="stonith" type="agent_type" id="r2"/>',
        ]
        self.assertEqual(len(expected_results), len(results))
        for i, res in enumerate(results):
            assert_xml_equal(expected_results[i], etree.tostring(res).decode())

    def test_with_provider(self):
        results = primitive._find_primitives_by_agent(
            self.resources_section,
            ResourceAgentName(
                "standard",
                "provider",
                "agent_type",
            ),
        )
        expected_results = [
            """<primitive
                class="standard" provider="provider" type="agent_type" id="r0"
            />""",
            """<primitive
                class="standard" provider="provider" type="agent_type" id="r4"
            />""",
            """<primitive
                class="standard" provider="provider" type="agent_type" id="r5"
            />""",
            """<primitive
                class="standard" provider="provider" type="agent_type" id="r6"
            />""",
        ]
        self.assertEqual(len(expected_results), len(results))
        for i, res in enumerate(results):
            assert_xml_equal(expected_results[i], etree.tostring(res).decode())


@mock.patch("pcs.lib.cib.resource.primitive.append_new_instance_attributes")
@mock.patch("pcs.lib.cib.resource.primitive.append_new_meta_attributes")
@mock.patch("pcs.lib.cib.resource.primitive.create_operations")
class AppendNew(TestCase):
    def setUp(self):
        self.resources_section = etree.fromstring("<resources/>")

        self.instance_attributes = {"a": "b"}
        self.meta_attributes = {"c": "d"}
        self.operation_list = [{"name": "monitoring"}]
        self.id_provider = IdProvider(self.resources_section)

        self.run = partial(
            primitive.append_new,
            self.resources_section,
            self.id_provider,
            instance_attributes=self.instance_attributes,
            meta_attributes=self.meta_attributes,
            operation_list=self.operation_list,
        )

    def check_mocks(
        self,
        primitive_element,
        create_operations,
        append_new_meta_attributes,
        append_new_instance_attributes,
    ):
        create_operations.assert_called_once_with(
            primitive_element, self.id_provider, self.operation_list
        )
        append_new_meta_attributes.assert_called_once_with(
            primitive_element, self.meta_attributes, self.id_provider
        )
        append_new_instance_attributes.assert_called_once_with(
            primitive_element, self.instance_attributes, self.id_provider
        )

    def test_append_without_provider(
        self,
        create_operations,
        append_new_meta_attributes,
        append_new_instance_attributes,
    ):
        primitive_element = self.run("RESOURCE_ID", "OCF", None, "DUMMY")
        self.assertEqual(
            primitive_element, self.resources_section.find(".//primitive")
        )
        self.assertEqual(primitive_element.attrib["class"], "OCF")
        self.assertEqual(primitive_element.attrib["type"], "DUMMY")
        self.assertFalse(primitive_element.attrib.has_key("provider"))

        self.check_mocks(
            primitive_element,
            create_operations,
            append_new_meta_attributes,
            append_new_instance_attributes,
        )

    def test_append_with_provider(
        self,
        create_operations,
        append_new_meta_attributes,
        append_new_instance_attributes,
    ):
        primitive_element = self.run("RESOURCE_ID", "OCF", "HEARTBEAT", "DUMMY")
        self.assertEqual(
            primitive_element, self.resources_section.find(".//primitive")
        )
        self.assertEqual(primitive_element.attrib["class"], "OCF")
        self.assertEqual(primitive_element.attrib["type"], "DUMMY")
        self.assertEqual(primitive_element.attrib["provider"], "HEARTBEAT")

        self.check_mocks(
            primitive_element,
            create_operations,
            append_new_meta_attributes,
            append_new_instance_attributes,
        )



from unittest import TestCase

from pcs.lib.cib.rule import tools
from pcs.lib.cib.rule.expression_part import (
    BOOL_AND,
    BOOL_OR,
    NODE_ATTR_OP_EQ,
    NODE_ATTR_TYPE_INTEGER,
    NODE_ATTR_TYPE_NUMBER,
    BoolExpr,
    NodeAttrExpr,
    OpExpr,
    RscExpr,
)


class HasNodeAttrExprWithTypeInteger(TestCase):
    @staticmethod
    def fixture_rule(attr_type):
        return BoolExpr(
            BOOL_OR,
            [
                BoolExpr(
                    BOOL_AND,
                    [
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "A", None),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "B", attr_type),
                    ],
                ),
                BoolExpr(
                    BOOL_AND,
                    [
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "aa", None),
                        RscExpr("ocf", "pacemaker", "Dummy"),
                    ],
                ),
                NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "1", None),
                OpExpr("stop", None),
            ],
        )

    def test_node_attr_no_type(self):
        self.assertFalse(
            tools.has_node_attr_expr_with_type_integer(self.fixture_rule(None))
        )

    def test_node_attr_no_integer(self):
        self.assertFalse(
            tools.has_node_attr_expr_with_type_integer(
                self.fixture_rule(NODE_ATTR_TYPE_NUMBER)
            )
        )

    def test_node_attr_integer(self):
        self.assertTrue(
            tools.has_node_attr_expr_with_type_integer(
                self.fixture_rule(NODE_ATTR_TYPE_INTEGER)
            )
        )

    def test_no_node_attr(self):
        self.assertFalse(
            tools.has_node_attr_expr_with_type_integer(
                BoolExpr(
                    BOOL_OR,
                    [
                        RscExpr("ocf", "pacemaker", "Dummy"),
                        OpExpr("stop", None),
                    ],
                )
            )
        )


class HasRscOrOpExpression(TestCase):
    def test_no_rsc_op(self):
        self.assertFalse(
            tools.has_rsc_or_op_expression(
                BoolExpr(
                    BOOL_OR,
                    [
                        BoolExpr(
                            BOOL_AND,
                            [
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "A", None),
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "B", None),
                            ],
                        ),
                        BoolExpr(
                            BOOL_AND,
                            [
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "aa", None),
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "bb", None),
                            ],
                        ),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "1", None),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "2", None),
                    ],
                )
            )
        )

    def test_rsc_on_top(self):
        self.assertTrue(
            tools.has_rsc_or_op_expression(RscExpr("ocf", "pacemaker", "Dummy"))
        )

    def test_op_on_top(self):
        self.assertTrue(tools.has_rsc_or_op_expression(OpExpr("stop", None)))

    def test_rsc_present(self):
        self.assertTrue(
            tools.has_rsc_or_op_expression(
                BoolExpr(
                    BOOL_OR,
                    [
                        BoolExpr(
                            BOOL_AND,
                            [
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "A", None),
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "B", None),
                            ],
                        ),
                        BoolExpr(
                            BOOL_AND,
                            [
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "aa", None),
                                RscExpr("ocf", "pacemaker", "Dummy"),
                            ],
                        ),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "1", None),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "2", None),
                    ],
                )
            )
        )

    def test_op_present(self):
        self.assertTrue(
            tools.has_rsc_or_op_expression(
                BoolExpr(
                    BOOL_OR,
                    [
                        BoolExpr(
                            BOOL_AND,
                            [
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "A", None),
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "B", None),
                            ],
                        ),
                        BoolExpr(
                            BOOL_AND,
                            [
                                OpExpr("stop", None),
                                NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "bb", None),
                            ],
                        ),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "a", "1", None),
                        NodeAttrExpr(NODE_ATTR_OP_EQ, "b", "2", None),
                    ],
                )
            )
        )



from textwrap import dedent


def fixture_corosync_conf_minimal(no_cluster_uuid=False):
    corosync_conf = """\
        totem {
            version: 2
            cluster_name: cluster_name
        """
    if not no_cluster_uuid:
        corosync_conf += """\
                cluster_uuid: cluster_uuid
            """

    corosync_conf += """\
            transport: knet
            ip_version: ipv6
            crypto_cipher: aes256
            crypto_hash: sha256
        }

        nodelist {
            node {
                ring0_addr: node1_addr
                name: node1
                nodeid: 1
            }

            node {
                ring0_addr: node2_addr
                name: node2
                nodeid: 2
            }
        }

        quorum {
            provider: corosync_votequorum
            two_node: 1
        }

        logging {
            to_logfile: yes
            logfile: /var/log/cluster/corosync.log
            to_syslog: yes
            timestamp: on
        }
        """

    return dedent(corosync_conf)



import os.path
import re
from dataclasses import dataclass
from typing import (
    Container,
    Optional,
    Sequence,
)

from pcs import settings
from pcs.common import reports
from pcs.common.file import RawFileError
from pcs.common.reports.item import ReportItem
from pcs.common.str_tools import format_list
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.raw_file import raw_file_error_report


def get_local_corosync_conf() -> str:
    """
    Read corosync.conf file from the local node
    """
    # TODO The architecture of working with corosync.conf needs to be
    # overhauled to match the new file framework.
    instance = FileInstance.for_corosync_conf()
    try:
        return instance.read_raw().decode("utf-8")
    except RawFileError as e:
        raise LibraryError(raw_file_error_report(e)) from e


def set_expected_votes(runner: CommandRunner, votes: int) -> str:
    """
    set expected votes in live cluster to the specified value
    """
    stdout, stderr, retval = runner.run(
        [
            os.path.join(settings.corosync_binaries, "corosync-quorumtool"),
            "-e",
            str(votes),
        ]
    )
    if retval != 0:
        raise LibraryError(
            ReportItem.error(
                reports.messages.CorosyncQuorumSetExpectedVotesError(stderr)
            )
        )
    return stdout


class QuorumStatusException(Exception):
    def __init__(self, reason: str = ""):
        super().__init__()
        self.reason = reason


class QuorumStatusReadException(QuorumStatusException):
    pass


class QuorumStatusParsingException(QuorumStatusException):
    pass


def get_quorum_status_text(runner: CommandRunner) -> str:
    """
    Get runtime quorum status from the local node
    """
    stdout, stderr, retval = runner.run(
        [os.path.join(settings.corosync_binaries, "corosync-quorumtool"), "-p"]
    )
    # retval is 0 on success if the node is not in a partition with quorum
    # retval is 1 on error OR on success if the node has quorum
    if retval not in [0, 1] or stderr.strip():
        raise QuorumStatusReadException(stderr)
    return stdout


@dataclass(frozen=True)
class QuorumStatusNode:
    name: str
    votes: int
    local: bool


@dataclass(frozen=True)
class QuorumStatus:
    node_list: Sequence[QuorumStatusNode]
    qdevice_list: Sequence[QuorumStatusNode]
    is_quorate: bool
    votes_needed_for_quorum: int

    @property
    def qdevice_votes(self) -> int:
        """
        How many votes are provided by qdevice(s)?
        """
        return sum(qdevice.votes for qdevice in self.qdevice_list)


class QuorumStatusFacade:
    @classmethod
    def from_string(cls, quorum_status_string: str) -> "QuorumStatusFacade":
        return cls(_parse_quorum_status(quorum_status_string))

    def __init__(self, quorum_status: QuorumStatus):
        self._quorum_status = quorum_status

    @property
    def node_list(self) -> Sequence[QuorumStatusNode]:
        return self._quorum_status.node_list

    @property
    def qdevice_list(self) -> Sequence[QuorumStatusNode]:
        return self._quorum_status.qdevice_list

    @property
    def is_quorate(self) -> bool:
        return self._quorum_status.is_quorate

    @property
    def votes_needed_for_quorum(self) -> int:
        return self._quorum_status.votes_needed_for_quorum

    @property
    def qdevice_votes(self) -> int:
        return self._quorum_status.qdevice_votes

    def _get_votes_excluding_nodes(self, node_names: Container[str]) -> int:
        """
        How many votes do remain if specified nodes are not counted in?

        node_names -- names of nodes not to count in
        """
        return (
            sum(
                node.votes
                for node in self.node_list
                if node.name not in node_names
            )
            + self.qdevice_votes
        )

    def _get_votes_excluding_local_node(self) -> int:
        """
        How many votes do remain if the local node is not counted in?
        """
        return (
            sum(node.votes for node in self.node_list if not node.local)
            + self.qdevice_votes
        )

    def stopping_nodes_cause_quorum_loss(
        self, node_names: Container[str]
    ) -> bool:
        """
        Will quorum be lost if specified nodes are stopped?

        node_names -- list of names of nodes to be stopped
        """
        if not self.is_quorate:
            return False
        return (
            self._get_votes_excluding_nodes(node_names)
            < self.votes_needed_for_quorum
        )

    def stopping_local_node_cause_quorum_loss(self) -> bool:
        """
        Will quorum be lost if the local node is stopped?
        """
        if not self.is_quorate:
            return False
        return (
            self._get_votes_excluding_local_node()
            < self.votes_needed_for_quorum
        )


def _parse_quorum_status(quorum_status: str) -> QuorumStatus:
    # pylint: disable=too-many-branches
    node_list: list[QuorumStatusNode] = []
    qdevice_list: list[QuorumStatusNode] = []
    quorate: Optional[bool] = None
    quorum: Optional[int] = None

    in_node_list = False
    try:
        for line in quorum_status.splitlines():
            line = line.strip()
            if not line:
                continue
            if in_node_list:
                if line.startswith("-") or line.startswith("Nodeid"):
                    # skip headers
                    continue
                parts = line.split()
                if parts[0] == "0":
                    # this line has nodeid == 0, this is a qdevice line
                    qdevice_list.append(
                        QuorumStatusNode(
                            name=parts[2],
                            votes=int(parts[1]),
                            local=False,
                        )
                    )
                else:
                    # this line has non-zero nodeid, this is a node line
                    node_list.append(
                        QuorumStatusNode(
                            name=parts[3],
                            votes=int(parts[1]),
                            local=(len(parts) > 4 and parts[4] == "(local)"),
                        )
                    )
            else:
                if line == "Membership information":
                    in_node_list = True
                    continue
                if not ":" in line:
                    continue
                parts = [x.strip() for x in line.split(":", 1)]
                if parts[0] == "Quorate":
                    quorate = parts[1].lower() == "yes"
                elif parts[0] == "Quorum":
                    match = re.match(r"(\d+).*", parts[1])
                    if match:
                        quorum = int(match.group(1))
                    else:
                        raise QuorumStatusParsingException(
                            "Unable to read number of votes needed for quorum"
                        )
    except (ValueError, IndexError) as e:
        raise QuorumStatusParsingException() from e
    missing_sections = []
    if quorum is None:
        missing_sections.append("quorum")
    if quorate is None:
        missing_sections.append("quorate")
    if not node_list:
        missing_sections.append("node_list")
    if missing_sections:
        raise QuorumStatusParsingException(
            "Missing required section(s): {}".format(
                format_list(missing_sections)
            )
        )
    return QuorumStatus(
        node_list=node_list,
        qdevice_list=qdevice_list,
        is_quorate=bool(quorate),
        # if ... else is here just for mypy, quorum can never be None, an
        # exception would be raised just above
        votes_needed_for_quorum=int(quorum) if quorum is not None else 0,
    )



import json
import re
from unittest import TestCase

from pcs import settings
from pcs.common import file_type_codes
from pcs.common.dr import DrRole
from pcs.common.file import RawFileError
from pcs.common.reports import codes as report_codes
from pcs.lib.commands import dr

from pcs_test.tools import fixture
from pcs_test.tools.command_env import get_env_tools

REASON = "error msg"


class CheckLive(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def assert_live_required(self, forbidden_options):
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env()),
            [
                fixture.error(
                    report_codes.LIVE_ENVIRONMENT_REQUIRED,
                    forbidden_options=forbidden_options,
                )
            ],
            expected_in_processor=False,
        )

    def test_mock_corosync(self):
        self.config.env.set_corosync_conf_data("corosync conf")
        self.assert_live_required([file_type_codes.COROSYNC_CONF])

    def test_mock_cib(self):
        self.config.env.set_cib_data("<cib />")
        self.assert_live_required([file_type_codes.CIB])

    def test_mock(self):
        self.config.env.set_corosync_conf_data("corosync conf")
        self.config.env.set_cib_data("<cib />")
        self.assert_live_required(
            [
                file_type_codes.CIB,
                file_type_codes.COROSYNC_CONF,
            ]
        )


class FixtureMixin:
    def _set_up(self, local_node_count=2):
        self.local_node_name_list = [
            f"node{i}" for i in range(1, local_node_count + 1)
        ]
        self.remote_node_name_list = ["recovery-node"]
        self.config.env.set_known_nodes(
            self.local_node_name_list + self.remote_node_name_list
        )
        self.local_status = "local cluster\nstatus\n"
        self.remote_status = "remote cluster\nstatus\n"

    def _fixture_load_configs(self):
        (
            self.config.raw_file.exists(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
            )
            .raw_file.read(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
                content="""
                    {
                        "local": {
                            "role": "PRIMARY"
                        },
                        "remote_sites": [
                            {
                                "nodes": [
                                    {
                                        "name": "recovery-node"
                                    }
                                ],
                                "role": "RECOVERY"
                            }
                        ]
                    }
                """,
            )
            .corosync_conf.load(node_name_list=self.local_node_name_list)
        )

    def _fixture_result(self, local_success=True, remote_success=True):
        return [
            {
                "local_site": True,
                "site_role": DrRole.PRIMARY,
                "status_plaintext": self.local_status if local_success else "",
                "status_successfully_obtained": local_success,
            },
            {
                "local_site": False,
                "site_role": DrRole.RECOVERY,
                "status_plaintext": (
                    self.remote_status if remote_success else ""
                ),
                "status_successfully_obtained": remote_success,
            },
        ]


class Success(FixtureMixin, TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self._set_up()

    def _assert_success(self, hide_inactive_resources, verbose):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                node_labels=self.local_node_name_list[:1],
                hide_inactive_resources=hide_inactive_resources,
                verbose=verbose,
                cluster_status_plaintext=self.local_status,
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                hide_inactive_resources=hide_inactive_resources,
                verbose=verbose,
                cluster_status_plaintext=self.remote_status,
            )
        )
        result = dr.status_all_sites_plaintext(
            self.env_assist.get_env(),
            hide_inactive_resources=hide_inactive_resources,
            verbose=verbose,
        )
        self.assertEqual(result, self._fixture_result())

    def test_success_minimal(self):
        self._assert_success(False, False)

    def test_success_full(self):
        self._assert_success(False, True)

    def test_success_hide_inactive(self):
        self._assert_success(True, False)

    def test_success_all_flags(self):
        self._assert_success(True, True)

    def test_local_not_running_first_node(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                cluster_status_plaintext=self.local_status,
                communication_list=[
                    [
                        dict(
                            label=self.local_node_name_list[0],
                            output=json.dumps(
                                dict(
                                    status="error",
                                    status_msg="",
                                    data=None,
                                    report_list=[
                                        {
                                            "severity": "ERROR",
                                            "code": "CRM_MON_ERROR",
                                            "info": {
                                                "reason": REASON,
                                            },
                                            "forceable": None,
                                            "report_text": "translated report",
                                        }
                                    ],
                                )
                            ),
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[1],
                        )
                    ],
                ],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cluster_status_plaintext=self.remote_status,
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result())
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=self.local_node_name_list[0],
                    command="remote/cluster_status_plaintext",
                    reason="translated report",
                ),
            ]
        )

    def test_local_not_running(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                cmd_status="error",
                cmd_status_msg="",
                cluster_status_plaintext="",
                report_list=[
                    {
                        "severity": "ERROR",
                        "code": "CRM_MON_ERROR",
                        "info": {
                            "reason": REASON,
                        },
                        "forceable": None,
                        "report_text": "translated report",
                    }
                ],
                communication_list=[
                    [
                        dict(
                            label=self.local_node_name_list[0],
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[1],
                        )
                    ],
                ],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cluster_status_plaintext=self.remote_status,
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result(local_success=False))
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=node,
                    command="remote/cluster_status_plaintext",
                    reason="translated report",
                )
                for node in self.local_node_name_list
            ]
        )

    def test_remote_not_running(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                node_labels=self.local_node_name_list[:1],
                cluster_status_plaintext=self.local_status,
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cmd_status="error",
                cmd_status_msg="",
                cluster_status_plaintext="",
                report_list=[
                    {
                        "severity": "ERROR",
                        "code": "CRM_MON_ERROR",
                        "info": {
                            "reason": REASON,
                        },
                        "forceable": None,
                        "report_text": "translated report",
                    }
                ],
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result(remote_success=False))
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=node,
                    command="remote/cluster_status_plaintext",
                    reason="translated report",
                )
                for node in self.remote_node_name_list
            ]
        )

    def test_both_not_running(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                cmd_status="error",
                cmd_status_msg="",
                cluster_status_plaintext="",
                report_list=[
                    {
                        "severity": "ERROR",
                        "code": "CRM_MON_ERROR",
                        "info": {
                            "reason": REASON,
                        },
                        "forceable": None,
                        "report_text": "translated report",
                    }
                ],
                communication_list=[
                    [
                        dict(
                            label=self.local_node_name_list[0],
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[1],
                        )
                    ],
                ],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cmd_status="error",
                cmd_status_msg="",
                cluster_status_plaintext="",
                report_list=[
                    {
                        "severity": "ERROR",
                        "code": "CRM_MON_ERROR",
                        "info": {
                            "reason": REASON,
                        },
                        "forceable": None,
                        "report_text": "translated report",
                    }
                ],
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(
            result,
            self._fixture_result(local_success=False, remote_success=False),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.NODE_COMMUNICATION_COMMAND_UNSUCCESSFUL,
                    node=node,
                    command="remote/cluster_status_plaintext",
                    reason="translated report",
                )
                for node in (
                    self.local_node_name_list + self.remote_node_name_list
                )
            ]
        )


class CommunicationIssue(FixtureMixin, TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)
        self._set_up()

    def test_unknown_node(self):
        self.config.env.set_known_nodes(
            self.local_node_name_list[1:] + self.remote_node_name_list
        )
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                node_labels=self.local_node_name_list[1:],
                cluster_status_plaintext=self.local_status,
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cluster_status_plaintext=self.remote_status,
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result())
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND,
                    host_list=["node1"],
                ),
            ]
        )

    def test_unknown_all_nodes_in_site(self):
        self.config.env.set_known_nodes(self.local_node_name_list)
        self._fixture_load_configs()
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env()),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.HOST_NOT_FOUND,
                    host_list=self.remote_node_name_list,
                ),
                fixture.error(
                    report_codes.NONE_HOST_FOUND,
                ),
            ]
        )

    def test_missing_node_names(self):
        self._fixture_load_configs()
        coro_call = self.config.calls.get("corosync_conf.load")
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                node_labels=[],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cluster_status_plaintext=self.remote_status,
            )
        )
        coro_call.content = re.sub(r"name: node\d", "", coro_call.content)
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result(local_success=False))
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.COROSYNC_CONFIG_MISSING_NAMES_OF_NODES,
                    fatal=False,
                ),
            ]
        )

    def test_node_issues(self):
        self._set_up(local_node_count=7)
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                cluster_status_plaintext=self.local_status,
                communication_list=[
                    [
                        dict(
                            label=self.local_node_name_list[0],
                            was_connected=False,
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[1],
                            response_code=401,
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[2],
                            response_code=500,
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[3],
                            response_code=404,
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[4],
                            output="invalid data",
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[5],
                            output=json.dumps(dict(status="success")),
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[6],
                        )
                    ],
                ],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cluster_status_plaintext=self.remote_status,
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result())
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="node1",
                    reason=None,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_NOT_AUTHORIZED,
                    command="remote/cluster_status_plaintext",
                    node="node2",
                    reason="HTTP error: 401",
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR,
                    command="remote/cluster_status_plaintext",
                    node="node3",
                    reason="HTTP error: 500",
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNSUPPORTED_COMMAND,
                    command="remote/cluster_status_plaintext",
                    node="node4",
                    reason="HTTP error: 404",
                ),
                fixture.warn(
                    report_codes.INVALID_RESPONSE_FORMAT,
                    node="node5",
                ),
                fixture.warn(
                    report_codes.INVALID_RESPONSE_FORMAT,
                    node="node6",
                ),
            ]
        )

    def test_local_site_down(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                cluster_status_plaintext=self.local_status,
                communication_list=[
                    [
                        dict(
                            label=self.local_node_name_list[0],
                            was_connected=False,
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[1],
                            was_connected=False,
                        )
                    ],
                ],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                node_labels=self.remote_node_name_list[:1],
                cluster_status_plaintext=self.remote_status,
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result(local_success=False))
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="node1",
                    reason=None,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="node2",
                    reason=None,
                ),
            ]
        )

    def test_remote_site_down(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                node_labels=self.local_node_name_list[:1],
                cluster_status_plaintext=self.local_status,
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                cluster_status_plaintext=self.remote_status,
                communication_list=[
                    [
                        dict(
                            label=self.remote_node_name_list[0],
                            was_connected=False,
                        )
                    ],
                ],
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(result, self._fixture_result(remote_success=False))
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="recovery-node",
                    reason=None,
                ),
            ]
        )

    def test_both_sites_down(self):
        self._fixture_load_configs()
        (
            self.config.http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.local",
                cluster_status_plaintext=self.local_status,
                communication_list=[
                    [
                        dict(
                            label=self.local_node_name_list[0],
                            was_connected=False,
                        )
                    ],
                    [
                        dict(
                            label=self.local_node_name_list[1],
                            was_connected=False,
                        )
                    ],
                ],
            ).http.status.get_full_cluster_status_plaintext(
                name="http.status.get_full_cluster_status_plaintext.remote",
                cluster_status_plaintext=self.remote_status,
                communication_list=[
                    [
                        dict(
                            label=self.remote_node_name_list[0],
                            was_connected=False,
                        )
                    ],
                ],
            )
        )
        result = dr.status_all_sites_plaintext(self.env_assist.get_env())
        self.assertEqual(
            result,
            self._fixture_result(local_success=False, remote_success=False),
        )
        self.env_assist.assert_reports(
            [
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="node1",
                    reason=None,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="node2",
                    reason=None,
                ),
                fixture.warn(
                    report_codes.NODE_COMMUNICATION_ERROR_UNABLE_TO_CONNECT,
                    command="remote/cluster_status_plaintext",
                    node="recovery-node",
                    reason=None,
                ),
            ]
        )


class FatalConfigIssue(TestCase):
    def setUp(self):
        self.env_assist, self.config = get_env_tools(self)

    def test_config_missing(self):
        (
            self.config.raw_file.exists(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
                exists=False,
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env()),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.DR_CONFIG_DOES_NOT_EXIST,
                ),
            ]
        )

    def test_config_read_error(self):
        (
            self.config.raw_file.exists(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
            ).raw_file.read(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
                exception_msg=REASON,
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env()),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.FILE_IO_ERROR,
                    file_type_code=file_type_codes.PCS_DR_CONFIG,
                    file_path=settings.pcsd_dr_config_location,
                    operation=RawFileError.ACTION_READ,
                    reason=REASON,
                ),
            ]
        )

    def test_config_parse_error(self):
        (
            self.config.raw_file.exists(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
            ).raw_file.read(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
                content="bad content",
            )
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env()),
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.PARSE_ERROR_JSON_FILE,
                    file_type_code=file_type_codes.PCS_DR_CONFIG,
                    file_path=settings.pcsd_dr_config_location,
                    line_number=1,
                    column_number=1,
                    position=0,
                    reason="Expecting value",
                    full_msg="Expecting value: line 1 column 1 (char 0)",
                ),
            ]
        )

    def test_corosync_conf_read_error(self):
        (
            self.config.raw_file.exists(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
            )
            .raw_file.read(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
                content="{}",
            )
            .corosync_conf.load_content("", exception_msg=REASON)
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env()),
            [
                fixture.error(
                    report_codes.FILE_IO_ERROR,
                    file_type_code=file_type_codes.COROSYNC_CONF,
                    operation=RawFileError.ACTION_READ,
                    reason=REASON,
                    file_path=settings.corosync_conf_file,
                ),
            ],
            expected_in_processor=False,
        )

    def test_corosync_conf_parse_error(self):
        (
            self.config.raw_file.exists(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
            )
            .raw_file.read(
                file_type_codes.PCS_DR_CONFIG,
                settings.pcsd_dr_config_location,
                content="{}",
            )
            .corosync_conf.load_content("wrong {\n  corosync")
        )
        self.env_assist.assert_raise_library_error(
            lambda: dr.status_all_sites_plaintext(self.env_assist.get_env())
        )
        self.env_assist.assert_reports(
            [
                fixture.error(
                    report_codes.PARSE_ERROR_COROSYNC_CONF_LINE_IS_NOT_SECTION_NOR_KEY_VALUE
                ),
            ]
        )



from pcs import (
    client,
    usage,
)
from pcs.cli.common.routing import create_router

client_cmd = create_router(
    {
        "help": lambda lib, argv, modifiers: print(usage.client(argv)),
        "local-auth": client.local_auth_cmd,
    },
    ["client"],
)



import logging
from typing import (
    Collection,
    Sequence,
    Set,
    cast,
)

from pcs.common.file import RawFileError
from pcs.lib.auth.const import (
    ADMIN_GROUP,
    SUPERUSER,
)
from pcs.lib.auth.types import AuthUser
from pcs.lib.file.instance import FileInstance
from pcs.lib.file.json import JsonParserException
from pcs.lib.interface.config import ParserErrorException

from .config.facade import FacadeV2
from .config.parser import ParserError
from .config.types import (
    ClusterPermissions,
    ConfigV2,
    PermissionAccessType,
    PermissionEntry,
    PermissionTargetType,
)


def _get_empty_facade(permissions: Sequence[PermissionEntry]) -> FacadeV2:
    return FacadeV2(
        ConfigV2(
            data_version=1,
            clusters=[],
            permissions=ClusterPermissions(
                local_cluster=permissions,
            ),
        )
    )


def _complete_access_list(
    access_list: Collection[PermissionAccessType],
) -> Set[PermissionAccessType]:
    if PermissionAccessType.SUPERUSER in access_list:
        return set(PermissionAccessType)
    if PermissionAccessType.FULL in access_list:
        return {
            PermissionAccessType.READ,
            PermissionAccessType.WRITE,
            PermissionAccessType.GRANT,
            PermissionAccessType.FULL,
        }
    new = set(access_list)
    if PermissionAccessType.WRITE in access_list:
        return new | {PermissionAccessType.READ}
    return new


class PermissionsChecker:
    def __init__(self, logger: logging.Logger) -> None:
        self._logger = logger
        self._config_file_instance = FileInstance.for_pcs_settings_config()

    def _get_facade(self) -> FacadeV2:
        if not self._config_file_instance.raw_file.exists():
            self._logger.debug(
                "File '%s' doesn't exist, using default configuration",
                self._config_file_instance.raw_file.metadata.path,
            )
            return _get_empty_facade(
                (
                    PermissionEntry(
                        type=PermissionTargetType.GROUP,
                        name=ADMIN_GROUP,
                        allow=(
                            PermissionAccessType.READ,
                            PermissionAccessType.WRITE,
                            PermissionAccessType.GRANT,
                        ),
                    ),
                )
            )
        try:
            return cast(FacadeV2, self._config_file_instance.read_to_facade())
        except ParserError as e:
            self._logger.error(
                "Unable to parse file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.msg,
            )
        except JsonParserException:
            self._logger.error(
                "Unable to parse file '%s': not valid json",
                self._config_file_instance.raw_file.metadata.path,
            )
        except ParserErrorException:
            self._logger.error(
                "Unable to parse file '%s'",
                self._config_file_instance.raw_file.metadata.path,
            )
        except RawFileError as e:
            self._logger.error(
                "Unable to read file '%s': %s",
                self._config_file_instance.raw_file.metadata.path,
                e.reason,
            )
        return _get_empty_facade(tuple())

    def get_permissions(self, auth_user: AuthUser) -> Set[PermissionAccessType]:
        if auth_user.username == SUPERUSER:
            return _complete_access_list((PermissionAccessType.SUPERUSER,))
        facade = self._get_facade()
        all_permissions: Set[PermissionAccessType] = set()
        for target_name, target_type in [
            (auth_user.username, PermissionTargetType.USER)
        ] + [(group, PermissionTargetType.GROUP) for group in auth_user.groups]:
            entry = facade.get_entry(target_name, target_type)
            if entry:
                all_permissions |= set(entry.allow)
        return _complete_access_list(all_permissions)

    def is_authorized(
        self, auth_user: AuthUser, access: PermissionAccessType
    ) -> bool:
        self._logger.debug(
            "Permission check: username=%s groups=%s access=%s",
            auth_user.username,
            ",".join(auth_user.groups),
            str(access.value),
        )
        user_permissions = self.get_permissions(auth_user)
        self._logger.debug(
            "Current user permissions: %s",
            ",".join(
                sorted(permission.value for permission in user_permissions)
            ),
        )
        result = access in user_permissions
        if result:
            self._logger.debug(
                "%s access granted", str(access.value).capitalize()
            )
        else:
            self._logger.debug(
                "%s access denied", str(access.value).capitalize()
            )
        return result



import xml.dom.minidom

from lxml import etree


def dom_get_child_elements(element):
    return [
        child
        for child in element.childNodes
        if child.nodeType == xml.dom.minidom.Node.ELEMENT_NODE
    ]


def etree_to_str(tree):
    # etree returns string in bytes: b'xml'
    # so there is bytes to str conversion
    return etree.tostring(tree, pretty_print=True).decode()


class XmlManipulation:
    @classmethod
    def from_file(cls, file_name):
        return cls(etree.parse(file_name).getroot())

    @classmethod
    def from_str(cls, string):
        return cls(etree.fromstring(string))

    def __init__(self, tree):
        self.tree = tree

    @staticmethod
    def __append_to_child(element, xml_string):
        element.append(etree.fromstring(xml_string))

    def append_to_first_tag_name(self, tag_name, *xml_string_list):
        for xml_string in xml_string_list:
            self.__append_to_child(
                self.tree.find(".//{0}".format(tag_name)), xml_string
            )
        return self

    def __str__(self):
        return etree_to_str(self.tree)


def get_xml_manipulation_creator_from_file(file_name):
    return lambda: XmlManipulation.from_file(file_name)



# pylint: disable=too-many-lines
from unittest import (
    TestCase,
    mock,
)

from lxml import etree

from pcs.common import reports
from pcs.lib.cib.resource import primitive
from pcs.lib.resource_agent import (
    ResourceAgentAction,
    ResourceAgentFacade,
    ResourceAgentMetadata,
    ResourceAgentName,
    ResourceAgentParameter,
    const,
)

from pcs_test.tools import fixture
from pcs_test.tools.assertions import assert_report_item_list_equal


def _fixture_metadata(name, parameters, actions=None, exists=True):
    return ResourceAgentMetadata(
        name=name,
        agent_exists=exists,
        ocf_version=const.OCF_1_0,
        shortdesc=None,
        longdesc=None,
        parameters=parameters,
        actions=actions or [],
    )


def _fixture_parameter(name, required, deprecated_by):
    return ResourceAgentParameter(
        name,
        shortdesc=None,
        longdesc=None,
        type="string",
        default=None,
        enum_values=None,
        required=required,
        advanced=False,
        deprecated=bool(deprecated_by),
        deprecated_by=deprecated_by,
        deprecated_desc=None,
        unique_group=None,
        reloadable=False,
    )


def _fixture_agent():
    return ResourceAgentFacade(
        _fixture_metadata(
            ResourceAgentName("standard", "provider", "type"),
            [
                _fixture_parameter("optional1_new", False, []),
                _fixture_parameter("optional1_old", False, ["optional1_new"]),
                _fixture_parameter("optional2_new", False, []),
                _fixture_parameter("optional2_old", False, ["optional2_new"]),
                _fixture_parameter("required1_new", True, []),
                _fixture_parameter("required1_old", True, ["required1_new"]),
                _fixture_parameter("required2_new", True, []),
                _fixture_parameter("required2_old", True, ["required2_new"]),
                _fixture_parameter("action", False, []),
            ],
        )
    )


def _fixture_agent_deprecated_loop():
    return ResourceAgentFacade(
        _fixture_metadata(
            ResourceAgentName("standard", "provider", "type"),
            [
                _fixture_parameter("loop1", True, ["loop1"]),
                _fixture_parameter("loop2a", True, ["loop2b"]),
                _fixture_parameter("loop2b", True, ["loop2a"]),
                _fixture_parameter("loop3a", True, ["loop3b"]),
                _fixture_parameter("loop3b", True, ["loop3c"]),
                _fixture_parameter("loop3c", True, ["loop3a"]),
            ],
        )
    )


def _fixture_stonith(actions=None):
    return ResourceAgentFacade(
        _fixture_metadata(
            ResourceAgentName("stonith", None, "type"),
            [
                _fixture_parameter("required", True, []),
                _fixture_parameter("optional", False, []),
                _fixture_parameter("action", True, []),
            ],
            actions=actions,
        )
    )


def _fixture_void(stonith=False):
    return ResourceAgentFacade(
        ResourceAgentMetadata(
            ResourceAgentName("stonith", None, "type")
            if stonith
            else ResourceAgentName("standard", "provider", "type"),
            agent_exists=False,
            ocf_version=const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=[],
            actions=[],
        )
    )


def _fixture_action(name):
    return ResourceAgentAction(
        name=name,
        timeout=None,
        interval=None,
        role=None,
        start_delay=None,
        depth=None,
        automatic=True,
        on_target=False,
    )


def _fixture_ocf_agent(exists=True, self_validation=True):
    actions = [_fixture_action("monitor")]
    if self_validation:
        actions.append(_fixture_action("validate-all"))
    return ResourceAgentFacade(
        _fixture_metadata(
            name=ResourceAgentName("ocf", "provider", "type"),
            parameters=[
                _fixture_parameter("required", True, []),
                _fixture_parameter("optional", False, []),
            ],
            actions=actions,
            exists=exists,
        )
    )


class ValidateResourceInstanceAttributesCreate(TestCase):
    def setUp(self):
        agent_self_validation_patcher = mock.patch(
            "pcs.lib.cib.resource.primitive.validate_resource_instance_attributes_via_pcmk"
        )
        self.addCleanup(agent_self_validation_patcher.stop)
        agent_self_validation_mock = agent_self_validation_patcher.start()
        self.addCleanup(agent_self_validation_mock.assert_not_called)
        self.cmd_runner = mock.Mock()

    def test_set_empty_string(self):
        options = [
            "required1_new",
            "required2_old",
            "optional1_new",
            "optional1_old",
            "unknown",
        ]
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {name: "" for name in options},
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name=name,
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                )
                for name in options
            ]
            + [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown"],
                    allowed=[
                        "action",
                        "optional1_new",
                        "optional1_old",
                        "optional2_new",
                        "optional2_old",
                        "required1_new",
                        "required1_old",
                        "required2_new",
                        "required2_old",
                    ],
                    option_type="resource",
                    allowed_patterns=[],
                ),
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="required2_old",
                    replaced_by=["required2_new"],
                ),
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="optional1_old",
                    replaced_by=["optional1_new"],
                ),
            ],
        )

    def test_set_all_required_params_one_deprecated_one_new(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_new": "A",
                    "required2_old": "B",
                },
                etree.Element("resources"),
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="required2_old",
                    replaced_by=["required2_new"],
                ),
            ],
        )

    def test_set_all_required_and_optional_params_one_deprecated_one_new(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_new": "A",
                    "required2_old": "B",
                    "optional1_new": "C",
                    "optional2_old": "D",
                },
                etree.Element("resources"),
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="required2_old",
                    replaced_by=["required2_new"],
                ),
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="optional2_old",
                    replaced_by=["optional2_new"],
                ),
            ],
        )

    def test_set_all_required_and_optional_params_both_deprecated_and_new(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_new": "A",
                    "required1_old": "B",
                    "required2_new": "C",
                    "required2_old": "D",
                    "optional1_new": "E",
                    "optional1_old": "F",
                    "optional2_new": "G",
                    "optional2_old": "H",
                },
                etree.Element("resources"),
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="required1_old",
                    replaced_by=["required1_new"],
                ),
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="required2_old",
                    replaced_by=["required2_new"],
                ),
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="optional1_old",
                    replaced_by=["optional1_new"],
                ),
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="optional2_old",
                    replaced_by=["optional2_new"],
                ),
            ],
        )

    def test_set_unknown_params(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_new": "A",
                    "required2_new": "B",
                    "unknown1": "C",
                    "unknown2": "D",
                },
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown1", "unknown2"],
                    allowed=[
                        "action",
                        "optional1_new",
                        "optional1_old",
                        "optional2_new",
                        "optional2_old",
                        "required1_new",
                        "required1_old",
                        "required2_new",
                        "required2_old",
                    ],
                    option_type="resource",
                    allowed_patterns=[],
                ),
            ],
        )

    def test_set_unknown_params_forced(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_new": "A",
                    "required2_new": "B",
                    "unknown1": "C",
                    "unknown2": "D",
                },
                etree.Element("resources"),
                force=True,
            ),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["unknown1", "unknown2"],
                    allowed=[
                        "action",
                        "optional1_new",
                        "optional1_old",
                        "optional2_new",
                        "optional2_old",
                        "required1_new",
                        "required1_old",
                        "required2_new",
                        "required2_old",
                    ],
                    option_type="resource",
                    allowed_patterns=[],
                ),
            ],
        )

    def test_missing_required(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {},
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="resource",
                    option_names=["required1_new", "required1_old"],
                    deprecated_names=["required1_old"],
                ),
                fixture.error(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="resource",
                    option_names=["required2_new", "required2_old"],
                    deprecated_names=["required2_old"],
                ),
            ],
        )

    def test_missing_required_forced(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {},
                etree.Element("resources"),
                force=True,
            ),
            [
                fixture.warn(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    option_type="resource",
                    option_names=["required1_new", "required1_old"],
                    deprecated_names=["required1_old"],
                ),
                fixture.warn(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    option_type="resource",
                    option_names=["required2_new", "required2_old"],
                    deprecated_names=["required2_old"],
                ),
            ],
        )

    def test_deprecation_loop(self):
        # Meta-data are broken - there are obsoleting loops. The point of the
        # test is to make sure pcs does not crash or loop forever. Error
        # reports are not that important, since meta-data are broken.
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent_deprecated_loop(),
                {
                    "loop3b": "value",
                },
                etree.Element("resources"),
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="loop3b",
                    replaced_by=["loop3c"],
                ),
            ],
        )

    def test_stonith_reports(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "unknown": "option",
                },
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown"],
                    allowed=["action", "optional", "required"],
                    option_type="stonith",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_names=["required"],
                ),
            ],
        )

    def test_resource_action_not_deprecated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "action": "reboot",
                    "required1_new": "value",
                    "required2_new": "value",
                },
                etree.Element("resources"),
            ),
            [],
        )

    def test_stonith_action_deprecated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "reboot",
                    "required": "value",
                },
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.DEPRECATED_OPTION,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )

    def test_stonith_action_deprecated_forced(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "reboot",
                    "required": "value",
                },
                etree.Element("resources"),
                force=True,
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )

    def test_stonith_action_empty(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "",
                    "required": "value",
                },
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="action",
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_stonith_action_not_set(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "required": "value",
                },
                etree.Element("resources"),
            ),
            [],
        )

    def test_void_checks_for_empty_strings(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_void(),
                {
                    "param1": "",
                    "param2": "value2",
                },
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="param1",
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
            ],
        )

    def test_void_stonith_check_for_action(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                _fixture_void(stonith=True),
                {
                    "param1": "",
                    "action": "reboot",
                    "param2": "value2",
                },
                etree.Element("resources"),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTION_VALUE,
                    option_name="param1",
                    option_value="",
                    allowed_values=None,
                    cannot_be_empty=True,
                    forbidden_characters=None,
                ),
                fixture.error(
                    reports.codes.DEPRECATED_OPTION,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )


class ValidateResourceInstanceAttributesCreateSelfValidation(TestCase):
    def setUp(self):
        agent_self_validation_patcher = mock.patch(
            "pcs.lib.cib.resource.primitive.validate_resource_instance_attributes_via_pcmk"
        )
        self.addCleanup(agent_self_validation_patcher.stop)
        self.agent_self_validation_mock = agent_self_validation_patcher.start()
        self.agent_self_validation_mock.return_value = True, []
        self.cmd_runner = mock.Mock()

    def test_disabled(self):
        attributes = {"required": "value"}
        facade = _fixture_ocf_agent()
        self.assertEqual(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=False,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_success(self):
        attributes = {"required": "value"}
        facade = _fixture_ocf_agent()
        self.assertEqual(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_called_once_with(
            self.cmd_runner,
            facade.metadata.name,
            attributes,
        )

    def test_force(self):
        attributes = {"required": "value"}
        facade = _fixture_ocf_agent()
        self.assertEqual(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=True,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_called_once_with(
            self.cmd_runner,
            facade.metadata.name,
            attributes,
        )

    def test_failure(self):
        attributes = {"required": "value"}
        facade = _fixture_ocf_agent()
        failure_reason = "failure reason"
        self.agent_self_validation_mock.return_value = False, failure_reason
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=True,
            ),
            [
                fixture.error(
                    reports.codes.AGENT_SELF_VALIDATION_RESULT,
                    result=failure_reason,
                    force_code=reports.codes.FORCE,
                )
            ],
        )
        self.agent_self_validation_mock.assert_called_once_with(
            self.cmd_runner,
            facade.metadata.name,
            attributes,
        )

    def test_stonith_check(self):
        attributes = {"required": "value"}
        facade = _fixture_stonith(actions=[_fixture_action("validate-all")])
        self.assertEqual(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_called_once_with(
            self.cmd_runner,
            facade.metadata.name,
            attributes,
        )

    def test_nonexisting_agent(self):
        attributes = {"required": "value"}
        facade = _fixture_ocf_agent(exists=False)
        self.assertEqual(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_provides_self_validation(self):
        attributes = {"required": "value"}
        facade = _fixture_ocf_agent(self_validation=False)
        self.assertEqual(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_previous_errors(self):
        attributes = {}
        facade = _fixture_ocf_agent()
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_create(
                self.cmd_runner,
                facade,
                attributes,
                etree.Element("resources"),
                force=False,
                enable_agent_self_validation=True,
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="resource",
                    option_names=["required"],
                ),
            ],
        )
        self.agent_self_validation_mock.assert_not_called()


class ValidateResourceInstanceAttributesUpdate(TestCase):
    _NAME = "a-resource"

    def setUp(self):
        agent_self_validation_patcher = mock.patch(
            "pcs.lib.cib.resource.primitive.validate_resource_instance_attributes_via_pcmk"
        )
        self.addCleanup(agent_self_validation_patcher.stop)
        agent_self_validation_mock = agent_self_validation_patcher.start()
        self.addCleanup(agent_self_validation_mock.assert_not_called)
        self.cmd_runner = mock.Mock()

    def _fixture_resources(self, parameters):
        resources_el = etree.Element("resources")
        primitive_el = etree.SubElement(
            resources_el, "primitive", dict(id=self._NAME)
        )
        nvset_el = etree.SubElement(primitive_el, "instance_attributes")
        for name, value in parameters.items():
            etree.SubElement(nvset_el, "nvpair", dict(name=name, value=value))
        return resources_el

    def test_remove_required_deprecated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required2_old": "",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="resource",
                    option_names=["required2_new", "required2_old"],
                    deprecated_names=["required2_old"],
                ),
            ],
        )

    def test_remove_required_new(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_new": "",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTION_OF_ALTERNATIVES_IS_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="resource",
                    option_names=["required1_new", "required1_old"],
                    deprecated_names=["required1_old"],
                ),
            ],
        )

    def test_remove_required_deprecated_set_new(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required2_old": "",
                    "required2_new": "B",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                    }
                ),
            ),
            [],
        )

    def test_remove_required_new_set_deprecated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "required1_old": "A",
                    "required1_new": "",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                    }
                ),
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="required1_old",
                    replaced_by=["required1_new"],
                ),
            ],
        )

    def test_set_optional(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "optional1_new": "A",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                    }
                ),
            ),
            [],
        )

    def test_remove_optional(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "optional1_new": "",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                    }
                ),
            ),
            [],
        )

    def test_dont_report_previously_missing_required(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "optional1_new": "A",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                    }
                ),
            ),
            [],
        )

    def test_set_unknown_params(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "unknown1": "C",
                    "unknown2": "D",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                        "unknown1": "c",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown2"],
                    allowed=[
                        "action",
                        "optional1_new",
                        "optional1_old",
                        "optional2_new",
                        "optional2_old",
                        "required1_new",
                        "required1_old",
                        "required2_new",
                        "required2_old",
                    ],
                    option_type="resource",
                    allowed_patterns=[],
                ),
            ],
        )

    def test_remove_unknown_params(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "unknown1": "",
                    "unknown2": "",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                        "unknown1": "C",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown2"],
                    allowed=[
                        "action",
                        "optional1_new",
                        "optional1_old",
                        "optional2_new",
                        "optional2_old",
                        "required1_new",
                        "required1_old",
                        "required2_new",
                        "required2_old",
                    ],
                    option_type="resource",
                    allowed_patterns=[],
                ),
            ],
        )

    def test_unknown_params_forced(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "unknown1": "",
                    "unknown2": "D",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "A",
                        "required2_old": "B",
                        "unknown1": "C",
                    }
                ),
                force=True,
            ),
            [
                fixture.warn(
                    reports.codes.INVALID_OPTIONS,
                    option_names=["unknown2"],
                    allowed=[
                        "action",
                        "optional1_new",
                        "optional1_old",
                        "optional2_new",
                        "optional2_old",
                        "required1_new",
                        "required1_old",
                        "required2_new",
                        "required2_old",
                    ],
                    option_type="resource",
                    allowed_patterns=[],
                ),
            ],
        )

    def test_deprecation_loop(self):
        # Meta-data are broken - there are obsoleting loops. The point of the
        # test is to make sure pcs does not crash or loop forever. Error
        # reports are not that important, since meta-data are broken.
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent_deprecated_loop(),
                {
                    "loop3b": "value",
                },
                self._NAME,
                self._fixture_resources({}),
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="resource",
                    option_name="loop3b",
                    replaced_by=["loop3c"],
                ),
            ],
        )

    def test_stonith_reports(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "required": "",
                    "unknown": "value",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required": "A",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.INVALID_OPTIONS,
                    force_code=reports.codes.FORCE,
                    option_names=["unknown"],
                    allowed=["action", "optional", "required"],
                    option_type="stonith",
                    allowed_patterns=[],
                ),
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_names=["required"],
                ),
            ],
        )

    def test_resource_action_not_deprecated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_agent(),
                {
                    "action": "reboot",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required1_new": "value",
                        "required2_new": "value",
                    }
                ),
            ),
            [],
        )

    def test_stonith_action_deprecated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "reboot",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "action": "offon",
                        "required": "value",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.DEPRECATED_OPTION,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )

    def test_stonith_action_deprecated_forced(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "reboot",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "action": "offon",
                        "required": "value",
                    }
                ),
                force=True,
            ),
            [
                fixture.warn(
                    reports.codes.DEPRECATED_OPTION,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )

    def test_stonith_action_add(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "reboot",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "required": "value",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.DEPRECATED_OPTION,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )

    def test_stonith_action_empty(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "action": "",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "action": "reboot",
                        "required": "value",
                    }
                ),
            ),
            [],
        )

    def test_stonith_action_not_updated(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_stonith(),
                {
                    "optional": "value",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "action": "reboot",
                        "required": "value",
                    }
                ),
            ),
            [],
        )

    def test_void_stonith_check_for_action(self):
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                _fixture_void(stonith=True),
                {
                    "param1": "",
                    "action": "reboot",
                    "param2": "value2",
                },
                self._NAME,
                self._fixture_resources(
                    {
                        "action": "reboot",
                        "required": "value",
                    }
                ),
            ),
            [
                fixture.error(
                    reports.codes.DEPRECATED_OPTION,
                    force_code=reports.codes.FORCE,
                    option_type="stonith",
                    option_name="action",
                    replaced_by=["pcmk_off_action", "pcmk_reboot_action"],
                ),
            ],
        )


class ValidateResourceInstanceAttributesUpdateSelfValidation(TestCase):
    _NAME = "a-resource"

    def setUp(self):
        agent_self_validation_patcher = mock.patch(
            "pcs.lib.cib.resource.primitive.validate_resource_instance_attributes_via_pcmk"
        )
        self.addCleanup(agent_self_validation_patcher.stop)
        self.agent_self_validation_mock = agent_self_validation_patcher.start()
        self.agent_self_validation_mock.return_value = True, []
        self.cmd_runner = mock.Mock()

    def _fixture_resources(self, parameters):
        resources_el = etree.Element("resources")
        primitive_el = etree.SubElement(
            resources_el, "primitive", dict(id=self._NAME)
        )
        nvset_el = etree.SubElement(primitive_el, "instance_attributes")
        for name, value in parameters.items():
            etree.SubElement(nvset_el, "nvpair", dict(name=name, value=value))
        return resources_el

    def test_disabled(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        facade = _fixture_ocf_agent()
        self.assertEqual(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=False,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_success(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        facade = _fixture_ocf_agent()
        self.assertEqual(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.assertEqual(
            self.agent_self_validation_mock.mock_calls,
            [
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    old_attributes,
                ),
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    new_attributes,
                ),
            ],
        )

    def test_force(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        facade = _fixture_ocf_agent()
        self.assertEqual(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=True,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.assertEqual(
            self.agent_self_validation_mock.mock_calls,
            [
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    old_attributes,
                ),
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    new_attributes,
                ),
            ],
        )

    def test_failure(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        failure_reason = "failure reason"
        facade = _fixture_ocf_agent()
        self.agent_self_validation_mock.side_effect = (
            (True, ""),
            (False, failure_reason),
        )
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [
                fixture.error(
                    reports.codes.AGENT_SELF_VALIDATION_RESULT,
                    result=failure_reason,
                    force_code=reports.codes.FORCE,
                )
            ],
        )
        self.assertEqual(
            self.agent_self_validation_mock.mock_calls,
            [
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    old_attributes,
                ),
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    new_attributes,
                ),
            ],
        )

    def test_stonith_check(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        facade = _fixture_stonith(actions=[_fixture_action("validate-all")])
        self.assertEqual(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.assertEqual(
            self.agent_self_validation_mock.mock_calls,
            [
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    old_attributes,
                ),
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    new_attributes,
                ),
            ],
        )

    def test_nonexisting_agent(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        facade = _fixture_ocf_agent(exists=False)
        self.assertEqual(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_provides_self_validation(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        facade = _fixture_ocf_agent(self_validation=False)
        self.assertEqual(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_previous_errors(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": ""}
        facade = _fixture_ocf_agent()
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [
                fixture.error(
                    reports.codes.REQUIRED_OPTIONS_ARE_MISSING,
                    force_code=reports.codes.FORCE,
                    option_type="resource",
                    option_names=["required"],
                ),
            ],
        )
        self.agent_self_validation_mock.assert_not_called()

    def test_current_attributes_failure(self):
        old_attributes = {"required": "old_value"}
        new_attributes = {"required": "new_value"}
        failure_reason = "failure reason"
        facade = _fixture_ocf_agent()
        self.agent_self_validation_mock.return_value = False, failure_reason
        assert_report_item_list_equal(
            primitive.validate_resource_instance_attributes_update(
                self.cmd_runner,
                facade,
                new_attributes,
                self._NAME,
                self._fixture_resources(old_attributes),
                force=False,
                enable_agent_self_validation=True,
            ),
            [
                fixture.warn(
                    reports.codes.AGENT_SELF_VALIDATION_SKIPPED_UPDATED_RESOURCE_MISCONFIGURED,
                    result=failure_reason,
                )
            ],
        )
        self.assertEqual(
            self.agent_self_validation_mock.mock_calls,
            [
                mock.call(
                    self.cmd_runner,
                    facade.metadata.name,
                    old_attributes,
                ),
            ],
        )


class ValidateUniqueInstanceAttributes(TestCase):
    # pylint: disable=protected-access
    cib = etree.fromstring(
        """
        <resources>
            <primitive class="ocf" provider="pacemaker" type="pcstest" id="R1">
                <instance_attributes>
                    <nvpair name="addr" value="127.0.0.1" />
                    <nvpair name="port" value="53" />
                    <nvapir name="something" value="else" />
                    <nvpair name="unique" value="value1"/>
                </instance_attributes>
            </primitive>
            <primitive class="ocf" provider="pacemaker" type="pcstest2" id="R2">
                <instance_attributes>
                    <nvpair name="addr" value="127.0.0.1" />
                    <nvpair name="port" value="53" />
                    <nvapir name="something" value="else" />
                    <nvpair name="unique" value="value1"/>
                </instance_attributes>
            </primitive>
            <clone id="G1-clone">
                <group id="G1">
                    <primitive class="ocf" provider="pacemaker" type="pcstest"
                        id="R3"
                    >
                        <instance_attributes>
                            <nvpair name="addr" value="127.0.0.1" />
                            <nvpair name="port" value="53" />
                            <nvapir name="something" value="else" />
                        </instance_attributes>
                    </primitive>
                </group>
            </clone>
        </resources>
        """
    )

    @staticmethod
    def _fixture_metadata():
        def _parameter(name, unique_group):
            return ResourceAgentParameter(
                name,
                shortdesc=None,
                longdesc=None,
                type="string",
                default=None,
                enum_values=None,
                required=False,
                advanced=False,
                deprecated=False,
                deprecated_by=None,
                deprecated_desc=None,
                unique_group=unique_group,
                reloadable=False,
            )

        return ResourceAgentMetadata(
            name=ResourceAgentName("ocf", "pacemaker", "pcstest"),
            agent_exists=True,
            ocf_version=const.OCF_1_0,
            shortdesc=None,
            longdesc=None,
            parameters=[
                _parameter("addr", "connection"),
                _parameter("port", "connection"),
                _parameter("something", None),
                _parameter("unique", "one-attr"),
            ],
            actions=[],
        )

    def test_no_report_on_different_values(self):
        assert_report_item_list_equal(
            primitive._validate_unique_instance_attributes(
                self._fixture_metadata(),
                {"addr": "127.0.0.2", "port": "54", "unique": "value2"},
                self.cib,
            ),
            [],
        )

    def test_no_report_when_not_all_values_in_group_are_same(self):
        assert_report_item_list_equal(
            primitive._validate_unique_instance_attributes(
                self._fixture_metadata(),
                {"addr": "127.0.0.2", "port": "53", "unique": "value2"},
                self.cib,
            ),
            [],
        )

    def test_report_same_values_from_same_agent_only(self):
        assert_report_item_list_equal(
            primitive._validate_unique_instance_attributes(
                self._fixture_metadata(),
                {"addr": "127.0.0.1", "port": "53", "unique": "value1"},
                self.cib,
            ),
            [
                fixture.error(
                    reports.codes.RESOURCE_INSTANCE_ATTR_GROUP_VALUE_NOT_UNIQUE,
                    force_code=reports.codes.FORCE,
                    group_name="connection",
                    instance_attrs_map={"addr": "127.0.0.1", "port": "53"},
                    agent_name="ocf:pacemaker:pcstest",
                    resource_id_list=["R1", "R3"],
                ),
                fixture.error(
                    reports.codes.RESOURCE_INSTANCE_ATTR_VALUE_NOT_UNIQUE,
                    force_code=reports.codes.FORCE,
                    instance_attr_name="unique",
                    instance_attr_value="value1",
                    agent_name="ocf:pacemaker:pcstest",
                    resource_id_list=["R1"],
                ),
            ],
        )

    def test_report_same_values_from_same_agent_only_forced(self):
        assert_report_item_list_equal(
            primitive._validate_unique_instance_attributes(
                self._fixture_metadata(),
                {"addr": "127.0.0.1", "port": "53", "unique": "value1"},
                self.cib,
                force=True,
            ),
            [
                fixture.warn(
                    reports.codes.RESOURCE_INSTANCE_ATTR_GROUP_VALUE_NOT_UNIQUE,
                    group_name="connection",
                    instance_attrs_map={"addr": "127.0.0.1", "port": "53"},
                    agent_name="ocf:pacemaker:pcstest",
                    resource_id_list=["R1", "R3"],
                ),
                fixture.warn(
                    reports.codes.RESOURCE_INSTANCE_ATTR_VALUE_NOT_UNIQUE,
                    instance_attr_name="unique",
                    instance_attr_value="value1",
                    agent_name="ocf:pacemaker:pcstest",
                    resource_id_list=["R1"],
                ),
            ],
        )

    def test_not_defined_values(self):
        assert_report_item_list_equal(
            primitive._validate_unique_instance_attributes(
                self._fixture_metadata(),
                {"addr": "127.0.0.2"},
                self.cib,
            ),
            [],
        )

    def test_ignore_own_values_on_update(self):
        assert_report_item_list_equal(
            primitive._validate_unique_instance_attributes(
                self._fixture_metadata(),
                {"unique": "value1"},
                self.cib,
                resource_id="R1",
            ),
            [],
        )



from typing import (
    Any,
    Dict,
    Iterable,
    List,
    Optional,
    cast,
)

from pcs.common.interface.dto import to_dict
from pcs.common.pacemaker.resource.operations import (
    OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME,
    ListCibResourceOperationDto,
)
from pcs.common.reports import (
    ReportItemSeverity,
    ReportProcessor,
)
from pcs.lib.cib.resource.agent import (
    get_default_operations,
    operation_dto_to_legacy_dict,
)
from pcs.lib.cib.resource.operations import uniquify_operations_intervals
from pcs.lib.cib.resource.types import ResourceOperationOut
from pcs.lib.env import LibraryEnvironment
from pcs.lib.errors import LibraryError
from pcs.lib.external import CommandRunner
from pcs.lib.resource_agent import (
    ListResourceAgentNameDto,
    ResourceAgentActionDto,
    ResourceAgentError,
    ResourceAgentFacadeFactory,
    ResourceAgentMetadata,
    ResourceAgentMetadataDto,
    ResourceAgentName,
    ResourceAgentNameDto,
    StandardProviderTuple,
    find_one_resource_agent_by_type,
    list_resource_agents,
    list_resource_agents_ocf_providers,
    list_resource_agents_standards,
    list_resource_agents_standards_and_providers,
    resource_agent_error_to_report_item,
    split_resource_agent_name,
)
from pcs.lib.resource_agent.name import name_to_void_metadata


def list_standards(lib_env: LibraryEnvironment) -> List[str]:
    """
    List resource agents standards (ocf, lsb, ... ) on the local host
    """
    return [
        standard
        for standard in list_resource_agents_standards(lib_env.cmd_runner())
        if standard != "stonith"
    ]


def list_ocf_providers(lib_env: LibraryEnvironment) -> List[str]:
    """
    List resource agents ocf providers on the local host
    """
    return list_resource_agents_ocf_providers(lib_env.cmd_runner())


def list_agents_for_standard_and_provider(
    lib_env: LibraryEnvironment, standard_provider: Optional[str] = None
) -> List[str]:
    """
    List resource agents for specified standard on the local host

    standard_provider -- standard[:provider], e.g. None, ocf, ocf:pacemaker
    """
    if standard_provider:
        if standard_provider[-1] == ":":
            standard_provider = standard_provider[:-1]
        std_prov_list = [
            StandardProviderTuple(*standard_provider.split(":", 1))
        ]
    else:
        std_prov_list = list_resource_agents_standards_and_providers(
            lib_env.cmd_runner()
        )
    agents = []
    for std_prov in std_prov_list:
        if std_prov.is_stonith:
            continue
        agents += list_resource_agents(lib_env.cmd_runner(), std_prov)
    return sorted(agents, key=str.lower)


# deprecated: use get_agent_list instead
# for now, it is transformed to a list of dicts for backward compatibility
def list_agents(
    lib_env: LibraryEnvironment,
    describe: bool = True,
    search: Optional[str] = None,
) -> List[Dict[str, Any]]:
    """
    List all resource agents on the local host, optionally filtered and
        described

    describe -- load and return agents' metadata as well
    search -- return only agents which name contains this string
    """
    runner = lib_env.cmd_runner()

    # list agents for all standards and providers
    agent_names = []
    for std_prov in list_resource_agents_standards_and_providers(runner):
        if std_prov.is_stonith:
            continue
        agent_names.extend(_get_agent_names(runner, std_prov))
    return _complete_agent_list(
        runner,
        lib_env.report_processor,
        sorted(agent_names, key=lambda item: item.full_name),
        describe,
        search,
    )


def _get_agent_names(
    runner: CommandRunner, standard_provider: StandardProviderTuple
) -> List[ResourceAgentName]:
    return [
        ResourceAgentName(
            standard_provider.standard, standard_provider.provider, agent
        )
        for agent in list_resource_agents(runner, standard_provider)
    ]


def get_agents_list(lib_env: LibraryEnvironment) -> ListResourceAgentNameDto:
    """
    List all resource agents on the local host
    """
    runner = lib_env.cmd_runner()
    agent_names = []
    for std_prov in list_resource_agents_standards_and_providers(runner):
        agent_names.extend(_get_agent_names(runner, std_prov))
    return ListResourceAgentNameDto(
        names=[
            name.to_dto()
            for name in sorted(agent_names, key=lambda item: item.full_name)
        ]
    )


def _action_to_operation(
    action: ResourceAgentActionDto,
) -> ResourceOperationOut:
    """
    Transform agent action data to CIB operation data
    """
    # This function bridges new agent framework, which provides data in
    # dataclasses, to old resource create code and transforms new data
    # structures to a format expected by the old code. When resource create is
    # overhauled, this function is expected to be removed.
    operation = {}
    for key, value in to_dict(action).items():
        if key == "depth":
            # "None" values are not put to CIB, so this keeps the key in place
            # while making sure it's not put in CIB. I'm not sure why depth ==
            # 0 is treated like this, but I keep it in place so the behavior is
            # the same as it has been for a long time. If pcs starts using
            # depth / OCF_CHECK_LEVEL or there is other demand for it, consider
            # changing this so value of "0" is put in CIB.
            operation[OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME] = (
                None if value == "0" else value
            )
        elif key == "start_delay":
            operation["start-delay"] = value
        else:
            operation[key] = value
    return cast(ResourceOperationOut, operation)


# backward compatibility layer - export agent metadata in the legacy format
def _agent_metadata_to_dict(
    agent: ResourceAgentMetadata, describe: bool = False
) -> Dict[str, Any]:
    agent_dto = agent.to_dto()
    agent_dict = to_dict(agent_dto)
    del agent_dict["name"]
    agent_dict["name"] = agent.name.full_name
    agent_dict["standard"] = agent.name.standard
    agent_dict["provider"] = agent.name.provider
    agent_dict["type"] = agent.name.type

    agent_dict["actions"] = [
        _action_to_operation(action) for action in agent_dto.actions
    ]
    operations_defaults = {
        OCF_CHECK_LEVEL_INSTANCE_ATTRIBUTE_NAME: None,
        "automatic": False,
        "on_target": False,
    }
    agent_dict["default_actions"] = (
        [
            operation_dto_to_legacy_dict(op, operations_defaults)
            for op in get_default_operations(agent)
        ]
        if describe
        else []
    )
    return agent_dict


def _complete_agent_list(
    runner: CommandRunner,
    report_processor: ReportProcessor,
    agent_names: Iterable[ResourceAgentName],
    describe: bool,
    search: Optional[str],
) -> List[Dict[str, Any]]:
    agent_factory = ResourceAgentFacadeFactory(runner, report_processor)
    search_lower = search.lower() if search else None
    agent_list = []
    for name in agent_names:
        if search_lower and search_lower not in name.full_name.lower():
            continue
        try:
            metadata = (
                agent_factory.facade_from_parsed_name(name).metadata
                if describe
                else name_to_void_metadata(name)
            )
            agent_list.append(_agent_metadata_to_dict(metadata, describe))
        except ResourceAgentError as e:
            report_processor.report(
                resource_agent_error_to_report_item(
                    e, ReportItemSeverity.warning()
                )
            )
    return agent_list


def _get_agent_metadata(
    runner: CommandRunner,
    report_processor: ReportProcessor,
    agent_name: ResourceAgentNameDto,
) -> ResourceAgentMetadata:
    agent_factory = ResourceAgentFacadeFactory(runner, report_processor)
    try:
        return agent_factory.facade_from_parsed_name(
            ResourceAgentName.from_dto(agent_name)
        ).metadata
    except ResourceAgentError as e:
        report_processor.report(resource_agent_error_to_report_item(e))
        raise LibraryError() from e


def get_agent_metadata(
    lib_env: LibraryEnvironment, agent_name: ResourceAgentNameDto
) -> ResourceAgentMetadataDto:
    """
    Return agent's metadata

    agent_name -- name of the agent
    """
    return _get_agent_metadata(
        lib_env.cmd_runner(),
        lib_env.report_processor,
        agent_name,
    ).to_dto()


# deprecated: use get_agent_metadata instead
# for now, it is transformed to a dict for backward compatibility
def describe_agent(
    lib_env: LibraryEnvironment, agent_name: str
) -> Dict[str, Any]:
    """
    Get agent's description (metadata) in a structure

    agent_name -- name of the agent
    """
    runner = lib_env.cmd_runner()
    report_processor = lib_env.report_processor
    agent_factory = ResourceAgentFacadeFactory(runner, report_processor)
    try:
        found_name = (
            split_resource_agent_name(agent_name)
            if ":" in agent_name
            else find_one_resource_agent_by_type(
                runner, report_processor, agent_name
            )
        )
        return _agent_metadata_to_dict(
            agent_factory.facade_from_parsed_name(found_name).metadata,
            describe=True,
        )
    except ResourceAgentError as e:
        lib_env.report_processor.report(resource_agent_error_to_report_item(e))
        raise LibraryError() from e


def get_agent_default_operations(
    lib_env: LibraryEnvironment,
    agent_name: ResourceAgentNameDto,
    necessary_only: bool = False,
) -> ListCibResourceOperationDto:
    report_list, operation_list = uniquify_operations_intervals(
        get_default_operations(
            _get_agent_metadata(
                lib_env.cmd_runner(), lib_env.report_processor, agent_name
            ),
            necessary_only,
        )
    )
    lib_env.report_processor.report_list(report_list)
    return ListCibResourceOperationDto(operations=operation_list)


def get_structured_agent_name(
    lib_env: LibraryEnvironment, agent_name: str
) -> ResourceAgentNameDto:
    # This is a temporary solution and should never be available via pcsd REST
    # API. The code for splitting an agent name will be eventually moved to cli
    # once all old commands that require this will be replaced and removed.
    try:
        return split_resource_agent_name(agent_name).to_dto()
    except ResourceAgentError as e:
        lib_env.report_processor.report(resource_agent_error_to_report_item(e))
        raise LibraryError() from e



from typing import (
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
)

from lxml.etree import _Element

from pcs.common import reports
from pcs.common.reports import (
    ReportItemList,
    ReportItemSeverity,
)
from pcs.common.reports.item import ReportItem
from pcs.lib.cib.node import (
    PacemakerNode,
    get_node_names,
)
from pcs.lib.cib.resource import (
    guest_node,
    remote_node,
)
from pcs.lib.corosync.config_facade import ConfigFacade as CorosyncConfigFacade
from pcs.lib.corosync.node import CorosyncNode
from pcs.lib.xml_tools import get_root


def get_existing_nodes_names(
    corosync_conf: Optional[CorosyncConfigFacade] = None,
    cib: Optional[_Element] = None,
    error_on_missing_name: bool = False,
) -> Tuple[List[str], ReportItemList]:
    return __get_nodes_names(
        *__get_nodes(corosync_conf, cib), error_on_missing_name
    )


def get_pacemaker_node_names(cib: _Element) -> Set[str]:
    return get_node_names(cib) | set(get_existing_nodes_names(None, cib)[0])


def get_existing_nodes_names_addrs(
    corosync_conf=None, cib=None, error_on_missing_name=False
):
    corosync_nodes, remote_and_guest_nodes = __get_nodes(corosync_conf, cib)
    names, report_list = __get_nodes_names(
        corosync_nodes, remote_and_guest_nodes, error_on_missing_name
    )
    return (
        names,
        __get_nodes_addrs(corosync_nodes, remote_and_guest_nodes),
        report_list,
    )


def __get_nodes(
    corosync_conf: Optional[CorosyncConfigFacade] = None,
    cib: Optional[_Element] = None,
) -> Tuple[Iterable[CorosyncNode], Iterable[PacemakerNode]]:
    corosync_nodes = corosync_conf.get_nodes() if corosync_conf else []
    remote_and_guest_nodes: Iterable[PacemakerNode] = []
    if cib is not None:
        cib_root = get_root(cib)
        remote_and_guest_nodes = remote_node.find_node_list(
            cib_root
        ) + guest_node.find_node_list(cib_root)
    return corosync_nodes, remote_and_guest_nodes


def __get_nodes_names(
    corosync_nodes: Iterable[CorosyncNode],
    remote_and_guest_nodes: Iterable[PacemakerNode],
    error_on_missing_name: bool = False,
) -> Tuple[List[str], ReportItemList]:
    report_list: ReportItemList = []
    corosync_names = []
    name_missing_in_corosync = False

    for node in corosync_nodes:
        if node.name:
            corosync_names.append(node.name)
        else:
            name_missing_in_corosync = True

    # Just a generic report for now. It may be expanded to provide more info
    # about each node missing a name later if needed.
    if name_missing_in_corosync:
        report_list.append(
            ReportItem(
                severity=(
                    ReportItemSeverity.error()
                    if error_on_missing_name
                    else ReportItemSeverity.warning()
                ),
                message=reports.messages.CorosyncConfigMissingNamesOfNodes(
                    fatal=error_on_missing_name,
                ),
            )
        )

    return (
        corosync_names + [node.name for node in remote_and_guest_nodes],
        report_list,
    )


def __get_nodes_addrs(corosync_nodes, remote_and_guest_nodes):
    nodes_addrs = [node.addr for node in remote_and_guest_nodes]
    for node in corosync_nodes:
        nodes_addrs += node.addrs_plain()
    return nodes_addrs
